#! /usr/bin/perl -w
#
# m2h.pl (Previously known as man2html)
# Copyright (C) 1997-2002 by Richard Dawe
#
# This software is distributed under the terms of the GNU General Public
# License, which should have been distributed with this file as LICENSE. Please
# read the notice below.
#
# ---
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
#
# ---
#
# Description:
#
# This program uses the 'man' command to generate text output which is then
# converted into HTML using the PRE pre-formatted tags. Any references like
# "socket(2)" are converted into hyperlinks.
#
# The program must be run from the directory below the 'man' tree, i.e. the
# man pages will be in the relative paths man/man1, man/man2, etc. The pages
# will be put into a similar tree with paths like html/html1, etc.
#
# The program also generates an alphabetic index file in the directory below
# the man tree called 'index.html' which has links to all the man pages
# converted.

# Uses
use strict;
no strict 'vars';
use FileHandle;
use DirHandle;

# Copyright message
$COPYRIGHT_AUTHOR = 'Richard Dawe';
$COPYRIGHT_AUTHOR_EMAIL = 'richdawe@bigfoot.com';

$COPYRIGHT_TEXT = 'm2h.pl Version 1.3.2, Copyright 1997-2002 by '
                  .$COPYRIGHT_AUTHOR.' <'.$COPYRIGHT_AUTHOR_EMAIL.'>';

$COPYRIGHT_MESSAGE = <<EOT;
$COPYRIGHT_TEXT

This program is distributed under the terms of the GNU Public License. Please
read the file license.txt that should have been distributed with this program.

EOT

# Creation message
$me =  $0;
$me =~ s?^(.*)/(.+)?$2?;
$CREATION_MESSAGE = '<I>Page created on '.localtime().' by '.$me
                    .', a man page to HTML convertor.<BR>'
                    .$COPYRIGHT_TEXT.'</I>';

$CREATION_MESSAGE =~ s!$COPYRIGHT_AUTHOR_EMAIL!<A HREF="mailto:$COPYRIGHT_AUTHOR_EMAIL">$COPYRIGHT_AUTHOR_EMAIL</A>!g;
$CREATION_MESSAGE =~ s!$COPYRIGHT_AUTHOR!<A HREF="mailto:$COPYRIGHT_AUTHOR_EMAIL">$COPYRIGHT_AUTHOR</A>!g;

# Help message
$HELP_MESSAGE = <<EOT;
m2h is a man page to HTML converter. It does this using a brute-force
approach - it *does not* understand the 'nroff' format of man pages. The
output is generated by filtering and search-and-replace.

The program expects the man pages to be in a directory structure like
man/man1, man/man2, etc. It creates directories like html/html1 (i.e.
analogous to the man directory names) with HTML-converted man pages in. The
program should be run from the directory in which the root man directory is
situated, e.g. /usr (e.g. /usr/man as the man root, /usr/html as the
HTML-converted man root).

Switches:
    -htm, --htm        => Output files will have the extension ".htm",
                          rather than ".html".

    -idx:<filename>,
    --index:<filename> => Create an index file called <filename>.

    -1, -one, --one,
    --all-in-ine       => Put all converted pages in one file.

    -sfn, --sfn        => Force filenames into MS-DOS's 8+3 filename limit.

Please see the m2h manual, m2h.htm, for more information.

EOT

$INDEX_FILE = 'index';          # Index filename stem
$HTML_EXTENSION = '.html';      # Long extension by default
$OUTPUT_ALLINONE = 0;           # Output all in one file?
$OUTPUT_SFN      = 0;           # Output with short filenames (8+3)?
# %HTML_LINK = {};

# Print the copyright message
print $COPYRIGHT_MESSAGE;

# Parse the command line arguments
for ($i = 0; $i <= $#ARGV; $i++) {
    # Help?
    if (   ($ARGV[$i] eq '-h')
        || ($ARGV[$i] eq '--help')
        || ($ARGV[$i] eq '-?') )
    {
        print $HELP_MESSAGE;
        exit;
    }

    # Index filename
    if ( ($ARGV[$i] =~ m/-idx:/) || ($ARGV[$i] =~ m/--index:/) ) {
        @temp = split(/\:/, $ARGV[$i]);
        $INDEX_FILE = $temp[1];
    }

    # .htm rather than .html?
    if ( ($ARGV[$i] eq '-htm') || ($ARGV[$i] eq '--htm') ) {
        $HTML_EXTENSION = '.htm';
    }

    if (   ($ARGV[$i] eq '-1')
        || ($ARGV[$i] eq '-one') || ($ARGV[$i] eq '--one')
        || ($ARGV[$i] eq '--all-in-one')
       ) {
        $OUTPUT_ALLINONE = 1;
    }

    # Short filenames?
    if ( ($ARGV[$i] eq '-sfn') || ($ARGV[$i] eq '--sfn') ) {
        $OUTPUT_SFN     = 1;
        $HTML_EXTENSION = '.htm';
    }
}

# Pages to put in index - format: caption, URL
@INDEXED = ();

# Check that the man directory exists
die 'No man directory in current directory' if (! -e 'man');
die 'man is not a directory' if (! -d 'man');

# Create the necessary html directories
if (! -e 'html')
{
    mkdir 'html', umask;
    print "Created \'html\' directory\n";
}

# Valid man<x> dirs, e.g. man1 - these two must be consistent for this program
# to work properly
@MANDIRS = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 'o', 'n', 'l', 'p');

# Recurse into these directories
chdir 'man';

for ($i = 0; $i <= $#MANDIRS; $i++)
{
    # Go to next one if this one doesn't exist
    if (! -e 'man' . $MANDIRS[$i])
    {
        print 'No man pages in section ' . $MANDIRS[$i] . "\n";
        next;
    }
    #else
    #{ print 'Converting section ' . $MANDIRS[$i] . "\n"; }

    # Create html dir if necessary
    if (! -e '../html/html' . $MANDIRS[$i])
    {
        if (!$OUTPUT_ALLINONE) {
            mkdir '../html/html' . $MANDIRS[$i], umask;
            print 'Created \'html/html' . $MANDIRS[$i] . "\' directory\n";
        }
    }

    # Find all the '.<section>' files & request them via man & save as HTML in
    # the HTML directory
    $man_dh = new DirHandle("man$MANDIRS[$i]");
    if (!defined($man_dh)) { die "Unable to open directory man$MANDIRS[$i]: $!"; }
    @FILES = $man_dh->read();
    $man_dh->close();

    for ($j = 0; $j <= $#FILES; $j++)
    {
	# Skip obvious directories, non-files
        if ( ($FILES[$j] eq '.') || ($FILES[$j] eq '..') ) { next; }	
	if (-d 'man'.$MANDIRS[$i].'/'.$FILES[$j]) { next; }

        $FILES[$j] = lc($FILES[$j]);        # Lowercase it for links
        $FILES[$j] =~ s/\.$MANDIRS[$i]//;  # Remove number extension

        $MANPAGENAME = $FILES[$j];
        #$MANPAGENAME =~ s/$\.$MANDIRS[$i]//;   # Done already

        # Build up the filename for the page
        if ($OUTPUT_SFN == 1) {
            $HTML_LINK{"$MANPAGENAME($MANDIRS[$i])"} = 'html'.$MANDIRS[$i].'/';
            if (length($FILES[$j]) > 8) {
                $HTML_LINK{"$MANPAGENAME($MANDIRS[$i])"} .= substr($FILES[$j], 0, 8);
            } else {
                $HTML_LINK{"$MANPAGENAME($MANDIRS[$i])"} .= $FILES[$j];
            }
            $HTML_LINK{"$MANPAGENAME($MANDIRS[$i])"} .= $HTML_EXTENSION;
        } else {
            $HTML_LINK{"$MANPAGENAME($MANDIRS[$i])"} =
                'html'.$MANDIRS[$i].'/'.$FILES[$j].$HTML_EXTENSION;
        }

        # Put entry in index
        #push(@INDEXED, ($MANPAGENAME.'('.$MANDIRS[$i].'):html'.$MANDIRS[$i].'/'.$FILES[$j].$HTML_EXTENSION));
        push(@INDEXED, (
                            "$MANPAGENAME($MANDIRS[$i]):"
                            . $HTML_LINK{"$MANPAGENAME($MANDIRS[$i])"}
                       ) );
    }
}

# Change back to parent directory of 'man' directory
chdir('..');

# Sort names of those pages to be converted
@INDEXED = sort(@INDEXED);

# Create the index file and converted pages in parallel
$index_fh = new FileHandle(">html/$INDEX_FILE$HTML_EXTENSION");
if (!defined($index_fh)) { die 'Unable to create index file'; }

$index_fh->print(
                   "<HTML>\n<HEAD>\n<TITLE>Man Page Index</TITLE>\n</HEAD>\n"
                 . "<BODY>\n<H1><A NAME=\"toc\">Man Page Index</A></H1>\n"
                );
              
for ($i = 'A'; $i lt 'Z'; $i++) { $index_fh->print("<A HREF=\"#$i\">$i</A> "); }
$index_fh->print("<A HREF=\"#Z\">Z</A><P>\n");

$letter = '';       # Used to generate targets for links at top
$manhtml = '';      # Converted man page
$manhtmlfile = '';  # Converted man page file name
$manref = '';       # $manpagename($mansection) style man references
$manpagename = '';  # man page name 
$mansection = '';   # man section

for ($i = 0; $i <= $#INDEXED; $i++)
{
    # Get details for the page to be indexed
    ($manref, $manhtmlfile) = split(/:/, $INDEXED[$i]);
    ($manpagename, $mansection) = split(/\(/, $manref);
    chop($mansection);

    # Do we need a new letter from the alphabet?
    if (uc(substr($manpagename, 0, 1)) ne $letter)
    {
        $letter = uc(substr($manpagename, 0, 1));
        $index_fh->print("<H2><A NAME=\"$letter\">$letter</A></H2>\n");
    }

    # If each man page has an HTML page, generate a link to its file, else
    # link to later in the page.
    if (!$OUTPUT_ALLINONE) {
        $index_fh->print(
                           '<A HREF="' . lc($manhtmlfile)
                         . "\">$manpagename($mansection)</A><BR>\n"
                        );
    } else {
        $index_fh->print(
                           '<A HREF="#' . $manpagename . $mansection
                         . "\">$manpagename($mansection)</A><BR>\n"
                        );
    }
}

# Divide between table of contents and rest in all-in-one style
if ($OUTPUT_ALLINONE) { $index_fh->print("<BR><HR>\n"); }

for ($i = 0; $i <= $#INDEXED; $i++)
{
    # Get details for the page to be indexed
    ($manref, $manhtmlfile) = split(/:/, $INDEXED[$i]);
    ($manpagename, $mansection) = split(/\(/, $manref);
    chop($mansection);

    # --- Generate the HTML for the man page ---

    # Open the output / use $index_fh
    if (!$OUTPUT_ALLINONE) {
        $manhtml_fh = new FileHandle(">html/$manhtmlfile");
    } else {
        $manhtml_fh = $index_fh;
    }

    if (!defined($manhtml_fh)) { die "Unable to open output '$manhtmlfile': $!"; }

    # Convert the man page
    $manhtml = &m2h($manpagename, $mansection, $HTML_EXTENSION, $OUTPUT_ALLINONE);

    if (!$OUTPUT_ALLINONE) {
        $manhtml_fh->print(
                             "<HTML>\n<HEAD>\n"
                           . "<TITLE>$manpagename($mansection)</TITLE>\n"
                           . "<BODY>\n" . $manhtml
                           . "<HR>\n<P>$CREATION_MESSAGE</P>"
                           . "</BODY>\n</HTML>"
                          );

        # Done now, so close it
        $manhtml_fh->close();
    } else {
        $manhtml_fh->print(
                             '<H3><A NAME="' . $manpagename . $mansection
                           . "\">$manpagename($mansection)</A></H3>\n"
                           . $manhtml
                          );

        $manhtml_fh->print(
                             "<P ALIGN=\"CENTER\">"
                           . "<A HREF=\"#toc\">[Top]</A>"
                           . "</P>\n<HR>\n"
                          );
    }
}

# Change the relative link for the translation issues file & output end of
# index
if (!$OUTPUT_ALLINONE) { $index_fh->print("<BR><HR>\n"); }
$index_fh->print("\n<P>$CREATION_MESSAGE</P>\n</BODY>\n</HTML>\n");
$index_fh->close();

# Finished
exit(0);

# -------
# - m2h -
# -------

# This fetches the required man page and converts it to HTML. The HTML does
# not include a header or body tags, so this can be used in the middle of an
# HTML document. The HTML is returned as a string.

sub m2h
{
    my ($manpagename, $mansection, $htmlext, $localref) = @_;
    my ($mandir, $manpage);
    my ($MANDIRS_REGEXP);

    # Used in link construction
    $MANDIRS_REGEXP = '0-9onlp';

    # Get the 'raw' output from man
    warn "Executing: man - $mansection $manpagename";
    $mandir = `pwd`;
    chomp($mandir);
    $mandir .= '/man';
    $manpage = `man -M $mandir - $mansection $manpagename`;

    # Nuke weird control characters
    $manpage =~ s/.\x08//g;

    # Escape characters
    $manpage =~ s!&!&amp;!g;
    $manpage =~ s!<!&lt;!g;
    $manpage =~ s!>!&gt;!g;

    # Generate links - if $localref == 1, then links to current document, else
    # to man pages in html/html<x>/ directory structure.
    if ($localref == 1) {
        $manpage =~ s/(\w+)\(([$MANDIRS_REGEXP])\)/\<A HREF\=\"\L#$1$2\E\"\>$1($2)\<\/A\>/g;
    } else {
        # ATTN: It lowercases in the regexp!
        #$manpage =~ s/(\w+)\(([$MANDIRS_REGEXP])\)/\<A HREF\=\"\L..\/html$2\/$1$htmlext\E\"\>$1($2)\<\/A\>/g;
        $manpage =~ s!(\w+)\(([$MANDIRS_REGEXP])\)!sprintf '%s', '<A HREF="../'.&manpage_html_link("$1($2)")."\">$1($2)</A>"!ge;
    }

    # Return it
    $manpage = '<P><PRE>' . $manpage . '</PRE></P>';    
    return($manpage);
}

# ---------------------
# - manpage_html_link -
# ---------------------

# Function to call in the second $manpage transformation above. This is so
# we can eliminate undefined references.

sub manpage_html_link {
    my ($key) = @_;
    my ($result);
    
    $key = lc($key);
    $result = '';
    if (defined $HTML_LINK{$key}) { $result = $HTML_LINK{$key}; }
    $result = lc($result);
    return($result);
}
